/*
Adafruit Trinket M0
SAMD21 + APA102

Bare-Metal playground for using a single APA102 with SERCOM and DMA
Done in the Arduino framework as the Trinket M0 has no SWD Header (just two solder pads).

Rudolph Riedel 02/2019
*/

#include "sam.h"

static DmacDescriptor dmadescriptor __attribute__((aligned(16))) SECTION_DMAC_DESCRIPTOR;
static DmacDescriptor dmawriteback __attribute__((aligned(16))) SECTION_DMAC_DESCRIPTOR;

#define DATA_LENGTH 12

uint8_t led_data[DATA_LENGTH] = {0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff};


void init_spi(void)
{
	/* configure SERCOM1 MOSI on PA00 and SERCOM1 SCK on PA01 */
	REG_PORT_WRCONFIG0 =
		PORT_WRCONFIG_WRPINCFG |
		PORT_WRCONFIG_WRPMUX |
		PORT_WRCONFIG_PMUX(3) |		/* SERCOM1 */
		PORT_WRCONFIG_DRVSTR |
		PORT_WRCONFIG_PINMASK(0x03) | /* PA00 + PA01 */
		PORT_WRCONFIG_PMUXEN;

	REG_PM_APBCMASK |= PM_APBCMASK_SERCOM1;
  REG_GCLK_CLKCTRL = GCLK_CLKCTRL_CLKEN | GCLK_CLKCTRL_ID_SERCOM1_CORE; /* setup SERCOM1 to use GLCK0 -> 48MHz */
	
	REG_SERCOM1_SPI_CTRLA = 0x00; /* disable SPI -> enable config */
	REG_SERCOM1_SPI_CTRLA = SERCOM_SPI_CTRLA_MODE(3) | SERCOM_SPI_CTRLA_DOPO(0); /* MSB first, CPOL = 0, CPHA = 0, SPI frame, master mode, PAD0 = MISO, PAD2 = MOSI, PAD3 = SCK */
	
  REG_SERCOM1_SPI_BAUD = 0; /* 48 / (2 * (baudval + 1)) -> @48Mhz: 0 = 24MHz, 5 = 4MHz, 23 = 1MHz, 47 = 500kHz */
	
  REG_SERCOM1_SPI_CTRLB = 0x00; /* receiver disabled, no hardware select, 8-bit */
	REG_SERCOM1_SPI_CTRLA |= SERCOM_SPI_CTRLA_ENABLE; /* activate SERCOM1 */
	while(REG_SERCOM1_SPI_SYNCBUSY & SERCOM_SPI_SYNCBUSY_ENABLE); /* wait for SERCOM1 to be ready */
}


void init_dma(void)
{
  REG_PM_APBBMASK |= PM_APBBMASK_DMAC;
  REG_PM_AHBMASK |= PM_AHBMASK_DMAC;

  REG_DMAC_BASEADDR = (uint32_t) &dmadescriptor;
  REG_DMAC_WRBADDR = (uint32_t) &dmawriteback;
  REG_DMAC_PRICTRL0 = 0; /* all off, reset-default */

  REG_DMAC_CHCTRLB = DMAC_CHCTRLB_TRIGACT_BEAT | DMAC_CHCTRLB_TRIGSRC(SERCOM1_DMAC_ID_TX); /* beat-transfer, SERCOM1 TX Trigger, level 0, channel-event input / output disabled */
  REG_DMAC_CTRL = DMAC_CTRL_LVLEN0 | DMAC_CTRL_DMAENABLE; /* enable level 0 transfers, enable DMA */
  REG_DMAC_CHID = 0; /* select channel 0, reset-default */

	dmadescriptor.BTCTRL.reg = DMAC_BTCTRL_SRCINC | DMAC_BTCTRL_VALID; /* increase source-address, beat-size = 8-bit */
  dmadescriptor.BTCNT.reg = DATA_LENGTH;
  dmadescriptor.SRCADDR.reg = (uint32_t) &led_data[DATA_LENGTH]; /* note: last entry in array + 1 */
  dmadescriptor.DSTADDR.reg = (uint32_t) &REG_SERCOM1_SPI_DATA;
  dmadescriptor.DESCADDR.reg = 0; /* no next descriptor */
}


/* write out data to APA102 LED as RGB */
void apa_dma_write(uint32_t color)
{
  if((REG_DMAC_CHCTRLA & DMAC_CHCTRLA_ENABLE) == 0)
  {
    led_data[5] = color; /* blue */
    led_data[6] = (color >> 8); /* green */
    led_data[7] = (color >> 16); /* red */
    REG_DMAC_CHCTRLA = DMAC_CHCTRLA_ENABLE;
  }
 }


void setup()
{
   init_spi();
   init_dma();
}


void loop()
{
  apa_dma_write(0x400000);
  delay(300);

  apa_dma_write(0x004000);
  delay(300);

  apa_dma_write(0x000040);
  delay(300);

  apa_dma_write(0x000000);
  delay(500);
}
